#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ncurses.h>

#define NAME_LEN 80
#define DEFAULT_PORT 87 
#define BUF_LEN 1024
#define TXWINHEIGHT 2

struct sockaddr_in myaddr, hisaddr;
void ttylink(int);

main(int argc, char *argv[])
{
   int sockfd, newsockfd, port, addrlen, childpid, wstatus;
   char c, myname[NAME_LEN], buff[BUF_LEN];
   char *pname = argv[0];

   if(argc < 2) port = DEFAULT_PORT;
   else if ((port = atoi(argv[1])) == 0) {
      fprintf(stderr, "Usage: %s [<port>]\n", argv[0]);
      exit(1);
   }

   system("clear");     /* clear screen */ 
   puts("Starting TTYLink server...");

   if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      fputs("TTYLink: Can't open socket", stderr);
      exit(1);
   }
   puts("Got a socket...");

   bzero((char *)&myaddr, sizeof(myaddr));
   myaddr.sin_family      = AF_INET;
   myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
   myaddr.sin_port        = htons(port);

   if (bind(sockfd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
      fputs("TTYLink: Can't bind local address", stderr);
      exit(1);
   }
   printf("Socket bound to TCP port %d...\n", port);  

   if (listen(sockfd, 1) == -1) {
      fputs("TTYLink: listen() error", stderr);
      exit(1);
   }
   puts("Listening for a connection...");


   for (;;) {
      addrlen = sizeof(hisaddr);
      newsockfd = accept(sockfd, (struct sockaddr *)&hisaddr, &addrlen);
      if (newsockfd < 0) {
         fputs("TTYLink: accept() error\n", stderr);
         exit(1);
      }
      
      if ( (childpid=fork()) < 0) {
         fputs("TTYLink: fork() error", stderr);
         exit(1);
      }
      else if (childpid == 0) { /* child process */
         close(sockfd);
         ttylink(newsockfd);      /* start session */
         exit(0);
      }

      /* Parent process */
      close(newsockfd);
      waitpid(childpid, &wstatus, 0); /* wait for current session to end */
      puts("TTYLink waiting for another connection...");
   }
}


/*
 * Most of the code here was borrowed from Dave Perry - VA3DP's 
 * implementation of the TTYLink client program.
 */
void ttylink(int sockfd)
{
   struct hostent *hp;
   int c = -1; 
   int recvretval;
   char rxbuff[BUF_LEN];
   char *rxptr = rxbuff;
   char *txptr, *txbuff;
   int txcount = 0;
   int rxcount = 0;
   struct timeval timeout;
   struct fd_set read_fdset;
   char myname[NAME_LEN];
   
   WINDOW *rxwin, *txwin;
   int x, y;
   
   initscr(); /* initialize ncurses */
   
   if ((txbuff = (char *)malloc(TXWINHEIGHT * COLS + 2)) == NULL) {
      fputs("TTYLink: malloc() failed\n", stderr);
      exit(1);
   }
   txptr = txbuff;
   
   refresh();   /* clear the screen */
   cbreak();    /* char mode input */
   noecho();    /* No automatic screen echo */

   rxwin = newwin(LINES-TXWINHEIGHT - 1, COLS, 0, 0); /* receive window */
   txwin = newwin(TXWINHEIGHT, COLS, LINES-TXWINHEIGHT-1, 0); /* xmit window */

   wattrset(rxwin, 0);
   wattrset(txwin, 0);
   scrollok(rxwin, TRUE);
   hp = gethostbyaddr((char*)&(hisaddr.sin_addr), 
                           sizeof(struct in_addr), AF_INET);

   beep();
   wprintw(rxwin, "*** Incoming connect from %s:%d\n\n", 
             hp->h_name, ntohs(hisaddr.sin_port) );
   wrefresh(rxwin);

   if (fcntl(sockfd, F_SETFL, FNDELAY) == -1) {  /* make socket non-blocking */
      fputs("TTYLink: fcntl() error on socket\n", stderr);
      exit(1);
   }


   gethostname(myname, NAME_LEN);  /* send welcome message */ 
   sprintf(txbuff, "Welcome to TTYLink for Linux at %s!\n\n", myname);
   send(sockfd, txbuff, strlen(txbuff), 0); 

   keypad(stdscr, TRUE); /* return single codes for function keys */

   for(;;) {
            /* now initialize file descriptor sets for select() */
      FD_ZERO(&read_fdset);
      FD_SET(STDIN_FILENO,&read_fdset);
      FD_SET(sockfd,&read_fdset);
      timeout.tv_sec = 30;      /* paranoid timer */
      timeout.tv_usec = 0;
      

      /* wake me up if something happens */
      /* For some reason, numfds has to be at least 5, even though */
      /* we've included only 2 fds in the fdset...                 */
      /* This is a man pages clarification opportunity... anybody? */
      if(select(5,&read_fdset,0,0,&timeout) == -1) {
         fputs("TTYLink: select() error\n", stderr);
         free(txbuff);
         exit(1);
      }

      /* get data from socket one byte at a time */
      while((recvretval = recv(sockfd, &c, 1, 0)) != -1) { 
         if (recvretval == 0) {   /* socket disconnected */
            echo();
            nocbreak();
            close(sockfd);
            endwin();
            puts("\007Connection closed by remote host...");
            free(txbuff);
            exit(0);
         }
         switch(c & 0xff) {
            case 0x07:   /* sound it, don't print it */
                      beep();
                      break;

            case '\r':   /* don't send these to ncurses */
                      break;

            case '\n':   /* complete line, display it */
                      *rxptr++ = c;
                      *rxptr = 0;
                      if(waddstr( rxwin, rxbuff) == -1)
                         puts("TTYLink: waddstr() failed");
                      rxptr = rxbuff;
                      rxcount = 0;
                      break;
         
            default:     /* save in rxbuff */
                      *rxptr++ = c;

                      if (rxcount++ == BUF_LEN - 1) { /* rx buffer full */
                         *rxptr = 0;                  /* display it     */
                          if (waddstr(rxwin, rxbuff) == -1)
                             puts("TTYLink: waddrstr() failed");
                          rxptr = rxbuff;
                          rxcount = 0;
                      }
                      break;

         } /* switch */
      }    /* while  */
      
      if (errno != EAGAIN) {
         echo();
         nocbreak();
         close(sockfd);
         endwin();
         free(txbuff);
         if (errno == ECONNRESET) {
           puts("\007Connection reset by peer");
           exit(0);
         }
         else {
            fprintf(stderr, "TTYLink: errno = %d\n", errno);
            exit(1);
         }
      }
    
      nodelay(stdscr, TRUE);  /* make getch() non-blocking */
      while((c = getch()) != -1) { 
         if ((c != KEY_BACKSPACE)) {
            if (!(c & 0xff00) && (c != '\n')) { /* ignore FN keys */
               if (++txcount <= (TXWINHEIGHT * COLS)) {
                  *txptr++ = c;
                  waddch(txwin, c);
               }
               else {
                  --txcount;
                  beep();
               }
            }
         }
         else {   /* backspace */
            if(txptr != txbuff) {
               txptr--;       /* delete it from the buffer */
               --txcount;

               getyx(txwin, y, x); /* and delete it from the screen */
               if (--x < 0) {
                  x = COLS - 1;
                  y--;
               }
               wmove(txwin, y, x);
               wdelch(txwin);
            }
         }
         if ((c & 0x7f) == '\n') {      /*  new line */
            *txptr++ = c;
            *txptr = 0;   /* null terminate */
            wclear(txwin);
            wattron(rxwin, A_BOLD);
            waddstr(rxwin, txbuff);
            wattroff(rxwin, A_BOLD);
            send(sockfd, txbuff, strlen(txbuff), 0);
            txptr = txbuff;
            txcount = 0;
         }
      } /* while */
 
      /* now update the screen */    
      nodelay(stdscr, FALSE);
      if(wrefresh(rxwin) == -1)
         puts("TTYLink: wrefresh() failed\n");
      if(wrefresh(txwin) == -1)
         puts("TTYLink: wrefresh() failed\n");

   } /* for(;;) */
}
